"""Exceptions raised by the emrserverless service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    code = 400

    def __init__(self, resource: str, resource_type: str = "Application"):
        super().__init__(
            "ResourceNotFoundException", f"{resource_type} {resource} does not exist"
        )


class ValidationException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class AccessDeniedException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("AccessDeniedException", message)



import json

import boto3

from moto import mock_aws


@mock_aws
def test_datapipeline():
    dp_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "dataPipeline": {
                "Properties": {
                    "Activate": "true",
                    "Name": "testDataPipeline",
                    "PipelineObjects": [
                        {
                            "Fields": [
                                {
                                    "Key": "failureAndRerunMode",
                                    "StringValue": "CASCADE",
                                },
                                {"Key": "scheduleType", "StringValue": "cron"},
                                {"Key": "schedule", "RefValue": "DefaultSchedule"},
                                {
                                    "Key": "pipelineLogUri",
                                    "StringValue": "s3://bucket/logs",
                                },
                                {"Key": "type", "StringValue": "Default"},
                            ],
                            "Id": "Default",
                            "Name": "Default",
                        },
                        {
                            "Fields": [
                                {
                                    "Key": "startDateTime",
                                    "StringValue": "1970-01-01T01:00:00",
                                },
                                {"Key": "period", "StringValue": "1 Day"},
                                {"Key": "type", "StringValue": "Schedule"},
                            ],
                            "Id": "DefaultSchedule",
                            "Name": "RunOnce",
                        },
                    ],
                    "PipelineTags": [],
                },
                "Type": "AWS::DataPipeline::Pipeline",
            }
        },
    }
    cf = boto3.client("cloudformation", region_name="us-east-1")
    template_json = json.dumps(dp_template)
    cf.create_stack(StackName="test_stack", TemplateBody=template_json)

    dp = boto3.client("datapipeline", region_name="us-east-1")
    data_pipelines = dp.list_pipelines()["pipelineIdList"]

    assert len(data_pipelines) == 1
    assert data_pipelines[0]["name"] == "testDataPipeline"

    stack_resources = cf.list_stack_resources(StackName="test_stack")[
        "StackResourceSummaries"
    ]
    assert len(stack_resources) == 1
    assert stack_resources[0]["PhysicalResourceId"] == data_pipelines[0]["id"]



"""quicksight base URL and path."""

from .responses import QuickSightResponse

url_bases = [
    r"https?://quicksight\.(.+)\.amazonaws\.com",
]


url_paths = {
    r"{0}/accounts/(?P<account_id>[\d]+)/data-sets$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/data-sets/(?P<datasetid>[^/.]+)/ingestions/(?P<ingestionid>[^/.]+)$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/groups$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/groups/(?P<groupname>[^/]+)$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/groups/(?P<groupname>[^/]+)/members$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/groups/(?P<groupname>[^/]+)/members/(?P<username>[^/]+)$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/users$": QuickSightResponse.dispatch,
    r"{0}/accounts/(?P<account_id>[\d]+)/namespaces/(?P<namespace>[a-zA-Z0-9._-]+)/users/(?P<username>[^/]+)$": QuickSightResponse.dispatch,
}



import abc
from collections import OrderedDict
from threading import Event
from typing import Dict, Final, Optional

from moto.moto_api._internal import mock_random

CallbackId = str


class CallbackOutcome(abc.ABC):
    callback_id: Final[CallbackId]

    def __init__(self, callback_id: str):
        self.callback_id = callback_id


class CallbackOutcomeSuccess(CallbackOutcome):
    output: Final[str]

    def __init__(self, callback_id: CallbackId, output: str):
        super().__init__(callback_id=callback_id)
        self.output = output


class CallbackOutcomeFailure(CallbackOutcome):
    error: Final[str]
    cause: Final[str]

    def __init__(self, callback_id: CallbackId, error: str, cause: str):
        super().__init__(callback_id=callback_id)
        self.error = error
        self.cause = cause


class CallbackTimeoutError(TimeoutError):
    pass


class CallbackConsumerError(abc.ABC): ...


class CallbackConsumerTimeout(CallbackConsumerError):
    pass


class CallbackConsumerLeft(CallbackConsumerError):
    pass


class HeartbeatEndpoint:
    _next_heartbeat_event: Final[Event]
    _heartbeat_seconds: Final[int]

    def __init__(self, heartbeat_seconds: int):
        self._next_heartbeat_event = Event()
        self._heartbeat_seconds = heartbeat_seconds

    def clear_and_wait(self) -> bool:
        self._next_heartbeat_event.clear()
        return self._next_heartbeat_event.wait(timeout=self._heartbeat_seconds)

    def notify(self):
        self._next_heartbeat_event.set()


class HeartbeatTimeoutError(TimeoutError):
    pass


class CallbackEndpoint:
    callback_id: Final[CallbackId]
    _notify_event: Final[Event]
    _outcome: Optional[CallbackOutcome]
    consumer_error: Optional[CallbackConsumerError]
    _heartbeat_endpoint: Optional[HeartbeatEndpoint]

    def __init__(self, callback_id: CallbackId):
        self.callback_id = callback_id
        self._notify_event = Event()
        self._outcome = None
        self.consumer_error = None
        self._heartbeat_endpoint = None

    def setup_heartbeat_endpoint(self, heartbeat_seconds: int) -> HeartbeatEndpoint:
        self._heartbeat_endpoint = HeartbeatEndpoint(
            heartbeat_seconds=heartbeat_seconds
        )
        return self._heartbeat_endpoint

    def notify(self, outcome: CallbackOutcome):
        self._outcome = outcome
        self._notify_event.set()
        if self._heartbeat_endpoint:
            self._heartbeat_endpoint.notify()

    def notify_heartbeat(self) -> bool:
        if not self._heartbeat_endpoint:
            return False
        self._heartbeat_endpoint.notify()
        return True

    def wait(self, timeout: Optional[float] = None) -> Optional[CallbackOutcome]:
        self._notify_event.wait(timeout=timeout)
        return self._outcome

    def get_outcome(self) -> Optional[CallbackOutcome]:
        return self._outcome

    def report(self, consumer_error: CallbackConsumerError) -> None:
        self.consumer_error = consumer_error


class CallbackNotifyConsumerError(RuntimeError):
    callback_consumer_error: CallbackConsumerError

    def __init__(self, callback_consumer_error: CallbackConsumerError):
        self.callback_consumer_error = callback_consumer_error


class CallbackOutcomeFailureError(RuntimeError):
    callback_outcome_failure: CallbackOutcomeFailure

    def __init__(self, callback_outcome_failure: CallbackOutcomeFailure):
        self.callback_outcome_failure = callback_outcome_failure


class CallbackPoolManager:
    _pool: Dict[CallbackId, CallbackEndpoint]

    def __init__(self):
        self._pool = OrderedDict()

    def get(self, callback_id: CallbackId) -> Optional[CallbackEndpoint]:
        return self._pool.get(callback_id)

    def add(self, callback_id: CallbackId) -> CallbackEndpoint:
        if callback_id in self._pool:
            raise ValueError("Duplicate callback token id value.")
        callback_endpoint = CallbackEndpoint(callback_id=callback_id)
        self._pool[callback_id] = callback_endpoint
        return callback_endpoint

    def generate(self) -> CallbackEndpoint:
        return self.add(str(mock_random.uuid4()))

    def notify(self, callback_id: CallbackId, outcome: CallbackOutcome) -> bool:
        callback_endpoint = self._pool.get(callback_id, None)
        if callback_endpoint is None:
            return False

        consumer_error: Optional[CallbackConsumerError] = (
            callback_endpoint.consumer_error
        )
        if consumer_error is not None:
            raise CallbackNotifyConsumerError(callback_consumer_error=consumer_error)

        callback_endpoint.notify(outcome=outcome)
        return True

    def heartbeat(self, callback_id: CallbackId) -> bool:
        callback_endpoint = self._pool.get(callback_id, None)
        if callback_endpoint is None:
            return False

        consumer_error: Optional[CallbackConsumerError] = (
            callback_endpoint.consumer_error
        )
        if consumer_error is not None:
            raise CallbackNotifyConsumerError(callback_consumer_error=consumer_error)

        return callback_endpoint.notify_heartbeat()



import pytest
from botocore.exceptions import ClientError

from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

from . import fixture_efs  # noqa # pylint: disable=unused-import


@pytest.fixture(scope="function", name="file_system")
def fixture_file_system(efs):
    create_fs_resp = efs.create_file_system(CreationToken="foobarbaz")
    create_fs_resp.pop("ResponseMetadata")
    yield create_fs_resp


def test_describe_access_points__initial(efs):
    resp = efs.describe_access_points()
    assert resp["AccessPoints"] == []


def test_create_access_point__simple(efs, file_system):
    fs_id = file_system["FileSystemId"]

    resp = efs.create_access_point(ClientToken="ct", FileSystemId=fs_id)
    assert resp["ClientToken"] == "ct"
    assert "AccessPointId" in resp
    assert "AccessPointArn" in resp
    assert resp["FileSystemId"] == fs_id
    assert resp["OwnerId"] == ACCOUNT_ID
    assert resp["LifeCycleState"] == "available"

    assert resp["RootDirectory"] == {"Path": "/"}


def test_create_access_point__full(efs, file_system):
    fs_id = file_system["FileSystemId"]

    resp = efs.create_access_point(
        ClientToken="ct",
        Tags=[{"Key": "key", "Value": "value"}, {"Key": "Name", "Value": "myname"}],
        FileSystemId=fs_id,
        PosixUser={"Uid": 123, "Gid": 123, "SecondaryGids": [124, 125]},
        RootDirectory={
            "Path": "/root/path",
            "CreationInfo": {
                "OwnerUid": 987,
                "OwnerGid": 986,
                "Permissions": "root_permissions",
            },
        },
    )

    assert resp["ClientToken"] == "ct"
    assert resp["Name"] == "myname"
    assert resp["Tags"] == [
        {"Key": "key", "Value": "value"},
        {"Key": "Name", "Value": "myname"},
    ]
    assert "AccessPointId" in resp
    assert "AccessPointArn" in resp
    assert resp["FileSystemId"] == fs_id
    assert resp["PosixUser"] == {"Uid": 123, "Gid": 123, "SecondaryGids": [124, 125]}
    assert resp["RootDirectory"] == {
        "Path": "/root/path",
        "CreationInfo": {
            "OwnerUid": 987,
            "OwnerGid": 986,
            "Permissions": "root_permissions",
        },
    }
    assert resp["OwnerId"] == ACCOUNT_ID
    assert resp["LifeCycleState"] == "available"


def test_describe_access_point(efs, file_system):
    fs_id = file_system["FileSystemId"]

    access_point_id = efs.create_access_point(ClientToken="ct", FileSystemId=fs_id)[
        "AccessPointId"
    ]

    resp = efs.describe_access_points(AccessPointId=access_point_id)
    assert len(resp["AccessPoints"]) == 1
    access_point = resp["AccessPoints"][0]

    assert access_point["ClientToken"] == "ct"
    assert "AccessPointId" in access_point
    assert "AccessPointArn" in access_point
    assert access_point["FileSystemId"] == fs_id
    assert access_point["OwnerId"] == ACCOUNT_ID
    assert access_point["LifeCycleState"] == "available"


def test_describe_access_points__multiple(efs, file_system):
    fs_id = file_system["FileSystemId"]

    efs.create_access_point(ClientToken="ct1", FileSystemId=fs_id)
    efs.create_access_point(ClientToken="ct2", FileSystemId=fs_id)

    resp = efs.describe_access_points()
    assert len(resp["AccessPoints"]) == 2


def test_describe_access_points__filters(efs):
    fs_id1 = efs.create_file_system(CreationToken="foobarbaz")["FileSystemId"]
    fs_id2 = efs.create_file_system(CreationToken="bazbarfoo")["FileSystemId"]

    ap_id1 = efs.create_access_point(ClientToken="ct", FileSystemId=fs_id1)[
        "AccessPointId"
    ]
    ap_id2 = efs.create_access_point(ClientToken="ct", FileSystemId=fs_id2)[
        "AccessPointId"
    ]

    resp = efs.describe_access_points(FileSystemId=fs_id1)
    assert len(resp["AccessPoints"]) == 1
    assert resp.get("NextToken") is None
    assert resp["AccessPoints"][0]["FileSystemId"] == fs_id1
    assert resp["AccessPoints"][0]["AccessPointId"] == ap_id1

    resp = efs.describe_access_points(AccessPointId=ap_id2)
    assert len(resp["AccessPoints"]) == 1
    assert resp.get("NextToken") is None
    assert resp["AccessPoints"][0]["FileSystemId"] == fs_id2
    assert resp["AccessPoints"][0]["AccessPointId"] == ap_id2


def test_delete_access_points(efs, file_system):
    fs_id = file_system["FileSystemId"]

    ap_id1 = efs.create_access_point(ClientToken="ct1", FileSystemId=fs_id)[
        "AccessPointId"
    ]
    ap_id2 = efs.create_access_point(ClientToken="ct2", FileSystemId=fs_id)[
        "AccessPointId"
    ]

    # Delete one access point
    efs.delete_access_point(AccessPointId=ap_id2)

    # We can only find one
    resp = efs.describe_access_points()
    assert len(resp["AccessPoints"]) == 1

    # The first one still exists
    efs.describe_access_points(AccessPointId=ap_id1)

    # The second one is gone
    with pytest.raises(ClientError) as exc_info:
        efs.describe_access_points(AccessPointId=ap_id2)
    err = exc_info.value.response["Error"]
    assert err["Code"] == "AccessPointNotFound"



import json

import moto.server as server


def test_appsync_list_tags_for_resource():
    backend = server.create_backend_app("appsync")
    test_client = backend.test_client()

    resp = test_client.get(
        "/v1/tags/arn%3Aaws%3Aappsync%3Aus-east-1%3A123456789012%3Aapis%2Ff405dd93-855e-451d-ab00-7325b8e439c6?tagKeys=Description"
    )
    assert resp.status_code == 200
    assert json.loads(resp.data) == {"tags": {}}



import boto3

from moto import mock_aws


@mock_aws
def test_disable_ebs_encryption_by_default():
    ec2 = boto3.client("ec2", "eu-central-1")

    ec2.enable_ebs_encryption_by_default()
    response = ec2.get_ebs_encryption_by_default()
    assert response["EbsEncryptionByDefault"] is True

    ec2.disable_ebs_encryption_by_default()
    after_disable_response = ec2.get_ebs_encryption_by_default()
    assert after_disable_response["EbsEncryptionByDefault"] is False


@mock_aws
def test_enable_ebs_encryption_by_default():
    ec2 = boto3.client("ec2", region_name="eu-central-1")
    response = ec2.enable_ebs_encryption_by_default()

    ec2.get_ebs_encryption_by_default()
    assert response["EbsEncryptionByDefault"] is True


@mock_aws
def test_get_ebs_encryption_by_default():
    ec2 = boto3.client("ec2", region_name="eu-west-1")

    response = ec2.get_ebs_encryption_by_default()
    assert response["EbsEncryptionByDefault"] is False


@mock_aws
def test_enable_ebs_encryption_by_default_region():
    ec2_eu = boto3.client("ec2", region_name="eu-central-1")
    ec2_eu.enable_ebs_encryption_by_default()

    response = ec2_eu.get_ebs_encryption_by_default()
    assert response["EbsEncryptionByDefault"] is True

    ec2_us = boto3.client("ec2", region_name="us-east-1")
    response = ec2_us.get_ebs_encryption_by_default()
    assert response["EbsEncryptionByDefault"] is False
